import numpy as np
import qiskit as qk
import pickle
from tqdm.notebook import tqdm
from copy import deepcopy

from .optimizers import *
from .layers import *
from .utils import *
from .samplers import *
from .costfunction import *


class NeuralNetwork():
    def __init__(self, layers=None, cost=MSE(), optimizer=Adam(lr=0.1)):
        self.layers = layers
        self.cost = cost
        self.optimizer = optimizer

        self.layers[0].last_layer = True
        self.dim = []

        if not self.layers == None:
            for layer in self.layers:
                self.dim.append(layer.weight.shape)

        if not self.optimizer == None:
            self.optimizer.initialize(self.dim)

        self.a = []
        self.weight_gradient_list = []

    def __call__(self, x, verbose=False):
        if verbose:
            decerator = tqdm
        else:
            decerator = identity

        self.a = []
        self.a.append(x)
        for layer in decerator(self.layers):
            x = layer(x)
            self.a.append(x)

    def predict(self, x, verbose=False):
        self(x, verbose=verbose)
        return self.a[-1]

    def backward(self, x, y=None, samplewise=False):
        n_samples = x.shape[0]
        self.weight_gradient_list = []

        self(x)
        y_pred = self.a[-1]

        delta = self.cost.derivative(y_pred, y)

        for i, layer in reversed(list(enumerate(self.layers))):
            weight_gradient, delta = layer.grad(
                self.a[i], delta, samplewise=samplewise)
            self.weight_gradient_list.append(weight_gradient)

        self.weight_gradient_list.reverse()

    def step(self):
        weight_gradient_modified = self.optimizer(self.weight_gradient_list)

        for layer, grad in zip(self.layers, weight_gradient_modified):
            layer.weight += -self.optimizer.lr * grad

    def train(self, x, y, epochs=100, verbose=False):
        if verbose:
            dec = tqdm
        else:
            dec = identity

        self.loss = []
        for i in dec(range(epochs)):

            self.backward(x, y)
            self.step()

            y_pred = self.a[-1]
            self.loss.append(np.mean((y_pred - y)**2))

            if verbose:
                print(f"epoch: {i}, loss: {self.loss[-1]}")

        y_pred = self.predict(x)
        self.loss.append(np.mean((y_pred - y)**2))

    def deriv(self, x):
        self.layers[0].last_layer = False
        self.weight_gradient_list = []

        self(x)
        delta = np.ones_like(x)

        for i, layer in reversed(list(enumerate(self.layers))):
            weight_gradient, delta = layer.grad(self.a[i], delta)

        self.layers[0].last_layer = True

        return delta

    @property
    def weight(self):
        weight_list = []
        for layer in self.layers:
            weight_list.append(layer.weight)

        return weight_list

    def randomize_weight(self):
        for layer in self.layers:
            layer.randomize_weight()

    def set_shots(self, shots):
        for layer in self.layers:
            layer.shots = shots

    def save(self, filename):
        pickle.dump(self, open(filename, "wb"))

    def load(self, filename):
        self = pickle.load(open(filename, "rb"))


def sequential_qnn(n_qubits=None,
                   dim=None,
                   encoder=Encoder(),
                   ansatz=Ansatz(reps=1),
                   sampler=LastBit(),
                   scale=None,
                   cost=MSE(),
                   optimizer=Adam(lr=0.1),
                   backend=None,
                   shots=None):
    L = len(dim)
    if scale == None:
        scale = (L - 2) * [[-np.pi, np.pi]]
        scale = scale + [1]

    layers = []
    for i in range(L - 1):
        in_dim = dim[i]
        out_dim = dim[i + 1]

        _encoder = deepcopy(encoder)
        _ansatz = deepcopy(ansatz)
        layer = QLayer(n_qubits=n_qubits[i],
                       n_features=in_dim,
                       n_targets=out_dim,
                       encoder=_encoder,
                       ansatz=_ansatz,
                       sampler=sampler,
                       scale=scale[i],
                       backend=backend,
                       shots=shots)
        layers.append(layer)

    network = NeuralNetwork(layers, cost=cost, optimizer=optimizer)

    return network


def sequential_dnn(dim=None,
                   bias=True,
                   scale=None,
                   cost=MSE(),
                   optimizer=Adam(lr=0.1)):
    L = len(dim)

    if scale == None:
        scale = (L - 1) * [1]

    layers = []
    for i in range(L - 1):
        in_dim = dim[i]
        out_dim = dim[i + 1]
        layer = Dense(n_features=in_dim, n_targets=out_dim, scale=scale[i],
                      activation=Sigmoid(), bias=bias)
        layers.append(layer)

    network = NeuralNetwork(layers, cost=cost, optimizer=optimizer)

    return network



import numpy as np
import qiskit as qk
import pickle
from tqdm.notebook import tqdm
from copy import deepcopy

from .optimizers import *
from .layers import *
from .utils import *
from .samplers import *
from .costfunction import *


class Ansatz():
    def __init__(self, blocks=["entangle", "ry", "rz"], reps=1):
        self.blocks = blocks
        self.reps = reps

    def __call__(self, circuit, data_register, weight, inverse=False):
        qregs = circuit.qregs[0]
        _circuit = qk.QuantumCircuit(qregs)

        idx_start = idx_end = 0
        for i in range(self.reps):
            for block in self.blocks:
                if block == "rx":
                    idx_end += self.n_qubits
                    for j, w in enumerate(weight[idx_start:idx_end]):
                        _circuit.rx(w, data_register[j])
                    idx_start = idx_end

                if block == "ry":
                    idx_end += self.n_qubits
                    for j, w in enumerate(weight[idx_start:idx_end]):
                        _circuit.ry(w, data_register[j])
                    idx_start = idx_end

                if block == "rz":
                    idx_end += self.n_qubits
                    for j, w in enumerate(weight[idx_start:idx_end]):
                        _circuit.rz(w, data_register[j])
                    idx_start = idx_end

                if block == "entangle":
                    for j in range(self.n_qubits - 1):
                        _circuit.cx(data_register[j], data_register[j + 1])

        if inverse:
            _circuit = _circuit.inverse()

        return circuit + _circuit

    def calculate_n_weights(self, n_qubits):
        self.n_qubits = n_qubits
        self.n_weights_per_target = 2 * self.reps * self.n_qubits


class RCO:

    def __init__(self, ansatz, n_qubits, sampler, optimizer, divisor, shots, tol=1e-3, warm_start=False):
        self.ansatz = ansatz
        self.sampler = sampler
        self.optimizer = optimizer
        self.divisor = divisor
        self.shots = shots
        self.tol = tol
        self.warm_start = warm_start
        self.n_qubits = n_qubits

        self.ansatz.calculate_n_weights(self.n_qubits)
        self.n_params = self.ansatz.n_weights_per_target
        self.params = np.random.uniform(-np.pi, np.pi,
                                        (self.divisor, self.n_params))

        self.error_sampler = ZeroBit()

    def fit(self, circuit):
        circuit_list = self.divide_circuit(circuit, self.divisor)
        params_prev = None
        print(f"{0}/{self.divisor} iterations")
        for i in range(self.divisor):
            if i != 0:
                params_prev = self.params[i - 1]
                if self.warm_start:
                    self.params[i] = np.copy(params_prev)

            self.optimize(circuit_list[i], self.params[i], params_prev)

            print(f"{i+1}/{self.divisor} iterations")

    def divide_circuit(self, circuit, divisor):
        circuit_size = len(circuit)
        gates_per_sub_circuit = circuit_size // divisor
        k = 0
        circuit_list = []
        while k < circuit_size:
            _circuit = deepcopy(circuit)
            for i in range(k):
                _circuit.data.pop(0)
            for i in range(circuit_size - gates_per_sub_circuit - k):
                _circuit.data.pop(-1)
            circuit_list.append(deepcopy(_circuit))
            k += gates_per_sub_circuit

        return circuit_list

    def evaluate(self, target_circuit, params, params_prev=None, return_circuit=False, use_error_measure=False):
        qreg = target_circuit.qregs[0]
        circuit = qk.QuantumCircuit(qreg)

        circuit = self.ansatz(circuit, qreg, params)
        circuit += target_circuit.inverse()
        if params_prev is not None:
            circuit = self.ansatz(circuit, qreg, params_prev, inverse=True)

        creg = qk.ClassicalRegister(qreg.size)
        circuit.add_register(creg)
        circuit.measure(qreg, creg)
        if return_circuit:
            return circuit
        else:
            job = qk.execute(circuit, backend=qk.Aer.get_backend(
                'qasm_simulator'), shots=self.shots)
            counts = job.result().get_counts(circuit)

            if use_error_measure == True:
                output = self.error_sampler(counts)
            else:
                output = self.sampler(counts)

            return output

    def gradient(self, target_circuit, params, params_prev=None):
        backend = qk.Aer.get_backend('qasm_simulator')

        grads = np.zeros(params.shape)
        circuit_plus_list = []
        circuit_minus_list = []
        for i in range(len(params)):
            params[i] += np.pi / 2
            circuit_plus_list.append(self.evaluate(
                target_circuit, params, params_prev, return_circuit=True))

            params[i] -= np.pi
            circuit_minus_list.append(self.evaluate(
                target_circuit, params, params_prev, return_circuit=True))

            params[i] += np.pi / 2

        circuit_list = circuit_plus_list + circuit_minus_list
        transpiled_list = qk.transpile(circuit_list, backend=backend)
        qobject_list = qk.assemble(transpiled_list,
                                   backend=backend,
                                   shots=self.shots,
                                   max_parallel_shots=0,
                                   max_parallel_experiments=0)

        job = backend.run(qobject_list)

        outputs = []
        for circuit in circuit_list:
            counts = job.result().get_counts(circuit)
            outputs.append(self.sampler(counts))

        outputs = np.array(outputs).reshape(2, -1)

        grads = (outputs[0] - outputs[1]) / 2
        return grads

    def optimize(self, target_circuit, params, params_prev=None):
        self.optimizer.initialize([len(params)])
        error = self.evaluate(target_circuit, params,
                              params_prev, use_error_measure=True)
        counter = 1
        while error > self.tol:
            grads = self.gradient(target_circuit, params, params_prev)
            grads = self.optimizer([grads])[0]

            params[:] = params - self.optimizer.lr * grads

            error = self.evaluate(target_circuit, params,
                                  params_prev, use_error_measure=True)

            print(f"{counter}: {error}")
            counter += 1

    def predict(self, target_circuit):
        return self.evaluate(target_circuit, self.params[-1], use_error_measure=True)



import numpy as np
import qiskit as qk
from copy import deepcopy
from tqdm.notebook import tqdm

from .neuralnetwork import *
from .utils import *


class FIM():
    """Empirical Fisher Information Matrix

    Parameters
    ----------
    model :
        Parameterized model
    """

    def __init__(self, model):
        self.model = model
        self.fim = None

    def fit(self, x):
        """Calculate the FIM for the given model over the input data x

        Parameters
        ----------
        x : ndarray
            input data
        """
        n_samples = x.shape[0]

        self.model.backward(x, samplewise=True)
        gradient = self.model.weight_gradient_list

        gradient_flattened = []
        for grad in gradient:
            gradient_flattened.append(grad.reshape(n_samples, -1))

        gradient_flattened = np.concatenate(gradient_flattened, axis=1)

        self.fim = 1 / n_samples * gradient_flattened.T @ gradient_flattened

    def eigen(self, sort=True):
        """Calculate the eigenvalue spectrum of the FIM.

        Parameters
        ----------
        sort : boolean
            Will sort the resulting eigenvalues in decending order if True.

        Notes
        -----
        Must be called after self.fit(x).
        """
        self.fim = np.array(self.fim, dtype=np.float64)
        self.eigen = np.linalg.eigh(self.fim)[0]

        if sort:
            self.eigen[::-1].sort()
        return np.maximum(self.eigen, 1e-25)

    def fisher_rao(self):
        """Calculate the Fisher-Rao metric.

        Notes
        -----
        Must be called after self.fit(x).
        """
        weight = self.model.weight

        weight_flattened = []
        for w in weight:
            weight_flattened.append(w.reshape(-1, 1))

        weight = np.concatenate(weight_flattened, axis=0)

        fr = weight.T @ self.fim @ weight

        return fr[0][0]


def trajectory_length(x):
    """Calculate the trajectory length of a discretized curve.

    Parameters
    ----------
    x : ndarray
        Discretized curve.
    """

    diff = (x[1:] - x[:-1])
    diff = np.append(diff, (x[0] - x[-1]).reshape(1, -1), axis=0)
    accum = np.sum(diff**2, axis=1)
    accum = np.sum(np.sqrt(accum))
    return accum


def trajectory_curvature(x):
    """Not implemented
    """
    pass


def gradient_analysis(network_list):
    n_models = len(network_list)
    gradients = np.zeros(
        (n_models, len(network_list[0].weight_gradient_list)))

    input_partial_avg = np.zeros(len(network_list[0].weight_gradient_list))
    weight_partial_avg = np.zeros(len(network_list[0].weight_gradient_list))

    for i, network in enumerate(network_list):
        for j in range(len(network.weight_gradient_list)):
            grad = network.weight_gradient_list[j]
            input_partial = network.layers[j].input_partial
            weight_partial = network.layers[j].weight_partial

            gradients[i, j] += np.mean(np.abs(grad))
            input_partial_avg[j] += np.mean(np.abs(input_partial))
            weight_partial_avg[j] += np.mean(np.abs(weight_partial))

    gradient_avg = np.mean(gradients, axis=0)
    gradient_std = np.std(gradients, axis=0)
    input_partial_avg /= n_models
    weight_partial_avg /= n_models

    return gradient_avg, input_partial_avg, weight_partial_avg



import numpy as np
import qiskit as qk
from math import ceil
from math import floor


class Parity():
    def __call__(self, counts):
        shots = sum(counts.values())
        output = 0
        for bitstring, samples in counts.items():
            if self.parity_of_bitstring(bitstring):
                output += samples

        output = output / shots

        return output

    def observable(self, circuit, data_register):
        return circuit

    def parity_of_bitstring(self, bitstring):
        binary = [int(i) for i in bitstring]
        parity = sum(binary) % 2

        return parity


class ZeroBit():
    def __call__(self, counts):
        shots = sum(counts.values())
        output = 0
        for bitstring, samples in counts.items():
            if "1" not in bitstring:
                output += samples

        output = output / shots

        return 1 - output


class AverageBit():
    def __call__(self, counts):
        shots = sum(counts.values())
        output = 0
        for bitstring, samples in counts.items():
            bits = [int(bit) for bit in bitstring]
            average = sum(bits) / len(bits)
            output += average * samples

        output = output / shots

        return output

    def observable(self, circuit, data_register):

        return circuit


class LastBit():
    def __call__(self, counts):
        shots = sum(counts.values())
        output = 0
        for bitstring, samples in counts.items():
            if bitstring[0] == "1":
                output += samples

        output = output / shots

        return output

    def observable(self, circuit, data_register):

        return circuit



import numpy as np


class MSE:
    def __call__(self, y_pred, y):
        return 0.5 * np.mean((y_pred - y)**2)

    def derivative(self, y_pred, y):
        return y_pred - y


class CrossEntropy:
    def __call__(self, y_pred, y):
        return -np.sum(y * np.log(y_pred))

    def derivative(self, y_pred, y):
        return (y_pred - y) / (y_pred * (1 - y_pred))


class NoCost:
    def __call__(self, y_pred, y):
        return y_pred

    def derivative(self, y_pred, y):
        n_samples, n_targets = y_pred.shape
        return np.ones((n_samples, n_targets))



import numpy as np
import qiskit as qk
from copy import deepcopy
from tqdm.notebook import tqdm

from .optimizers import *
from .data_encoders import *
from .samplers import *
from .utils import *


class Ansatz():
    def __init__(self, blocks=["entangle", "ry"], reps=1):
        self.blocks = blocks
        self.reps = reps

    def __call__(self, circuit, data_register, weight):

        idx_start = idx_end = 0
        for i in range(self.reps):
            for block in self.blocks:
                if block == "rx":
                    idx_end += self.n_qubits
                    for j, w in enumerate(weight[idx_start:idx_end]):
                        circuit.rx(w, data_register[j])
                    idx_start = idx_end

                if block == "ry":
                    idx_end += self.n_qubits
                    for j, w in enumerate(weight[idx_start:idx_end]):
                        circuit.ry(w, data_register[j])
                    idx_start = idx_end

                if block == "rz":
                    idx_end += self.n_qubits
                    for j, w in enumerate(weight[idx_start:idx_end]):
                        circuit.rz(w, data_register[j])
                    idx_start = idx_end

                if block == "entangle":
                    for j in range(self.n_qubits - 1):
                        circuit.cx(data_register[j], data_register[j + 1])

        return circuit

    def calculate_n_weights(self, n_qubits):
        self.n_qubits = n_qubits
        self.n_weights_per_target = self.reps * self.n_qubits


class Ansatz2():
    def __call__(self, circuit, data_register, weight):
        n_qubits = data_register.size

        for i, w in enumerate(weight[:n_qubits // 2]):
            circuit.cx(data_register[2 * i], data_register[2 * i + 1])
            circuit.rz(w, data_register[i])
            circuit.cx(data_register[2 * i], data_register[2 * i + 1])

        for i, w in enumerate(weight[n_qubits // 2:]):
            circuit.cx(data_register[2 * i + 1], data_register[2 * i + 2])
            circuit.rz(w, data_register[i])
            circuit.cx(data_register[2 * i + 1], data_register[2 * i + 2])

        return circuit



import numpy as np


class GD():
    def __init__(self, lr=0.01):
        self.lr = lr

    def initialize(self, dims):
        pass

    def __call__(self, weight_gradient_list):
        return weight_gradient_list


class Adam():
    def __init__(self, lr=0.01, beta1=0.9, beta2=0.999, eps=1e-8):
        self.lr = lr
        self.beta1 = beta1
        self.beta2 = beta2
        self.eps = eps
        self.m = None
        self.v = None
        self.t = None

    def initialize(self, dims):
        self.m = []
        self.v = []
        self.t = 0

        for dim in dims:
            self.m.append(np.zeros(dim))
            self.v.append(np.zeros(dim))

    def __call__(self, weight_gradient_list):
        self.t += 1
        weight_gradient_modified = []

        for grad, m_, v_ in zip(weight_gradient_list, self.m, self.v):
            m_[:] = self.beta1 * m_ + (1 - self.beta1) * grad
            v_[:] = self.beta2 * v_ + (1 - self.beta2) * grad**2

            m_hat = m_ / (1 - self.beta1**self.t)
            v_hat = v_ / (1 - self.beta2**self.t)
            grad_modified = m_hat / (np.sqrt(v_hat) + self.eps)
            weight_gradient_modified.append(grad_modified)

        return weight_gradient_modified



import numpy as np
import qiskit as qk
from copy import deepcopy

from .optimizers import *
from .data_encoders import *
from .ansatzes import *


class Dense():
    def __init__(self, n_features=None, n_targets=None, scale=1, activation=None, bias=True):

        self.n_features = n_features
        self.n_targets = n_targets
        self.scale = scale
        self.activation = activation
        self.bias = bias

        self.randomize_weight()

    def __call__(self, inputs):
        x = inputs @ self.weight[:self.n_features]
        if self.bias:
            x += self.weight[-1].reshape(1, -1)

        x = self.activation(x)

        return self.scale * x

    def grad(self, inputs, delta, samplewise=False):
        n_samples = inputs.shape[0]
        output = self(inputs)
        delta = self.scale * \
            self.activation.derivative(output / self.scale) * delta

        if not samplewise:
            weight_gradient = 1 / n_samples * inputs.T @ delta
        else:
            weight_gradient = [np.outer(input_, delta_)
                               for input_, delta_ in zip(inputs, delta)]
            weight_gradient = np.array(weight_gradient)

        if self.bias:
            if not samplewise:
                bias_gradient = np.mean(delta, axis=0, keepdims=True)

                weight_gradient = np.concatenate(
                    (weight_gradient, bias_gradient), axis=0)
            else:
                bias_gradient = delta

                weight_gradient = np.concatenate(
                    (weight_gradient, bias_gradient[:, None, :]), axis=1)

        delta = delta @ self.weight[:self.n_features].T

        return weight_gradient, delta

    def randomize_weight(self):
        """Xavier initialization"""

        std = 1 / np.sqrt(self.n_features)
        self.weight = np.random.uniform(
            -std, std, (self.n_features + self.bias, self.n_targets))


class QLayer():
    def __init__(self,
                 n_qubits=None,
                 n_features=None,
                 n_targets=None,
                 scale=1,
                 encoder=None,
                 ansatz=None,
                 sampler=None,
                 backend=None,
                 shots=1000):

        self.n_qubits = n_qubits
        self.n_features = n_features
        self.n_targets = n_targets
        self.scale = scale
        self.encoder = encoder
        self.ansatz = ansatz
        self.sampler = sampler
        self.backend = backend
        self.shots = shots

        self.encoder.calculate_n_weights(self.n_qubits)
        self.ansatz.calculate_n_weights(self.n_qubits)
        self.n_weights_per_target = self.encoder.n_weights_per_target + \
            self.ansatz.n_weights_per_target

        self.last_layer = False
        self.randomize_weight()

    def __call__(self, inputs):
        backend = qk.providers.aer.StatevectorSimulator()
        outputs = []
        circuit_list = []
        n_samples = inputs.shape[0]
        for x in inputs:
            for i in range(self.n_targets):
                data_register = qk.QuantumRegister(
                    self.n_qubits, name="storage")
                clas_register = qk.ClassicalRegister(
                    self.n_qubits, name="clas_reg")
                registers = [data_register]

                if self.shots != 0:
                    registers.append(clas_register)

                circuit = qk.QuantumCircuit(*registers)

                idx = self.encoder.n_weights_per_target
                self.encoder(circuit, data_register,
                             self.weight[:idx, i], x)

                self.ansatz(circuit, data_register,
                            self.weight[idx:, i])

                if self.shots == 0:
                    circuit = self.sampler.observable(circuit, data_register)
                else:
                    circuit.measure(data_register, clas_register)

                circuit_list.append(circuit)

        #transpiled_list = qk.transpile(circuit_list, backend=self.backend)

        if self.shots == 0:
            for circuit in circuit_list:
                counts = qk.execute(circuit, backend).result().get_counts()
                outputs.append(self.sampler(counts))
        else:
            qobject_list = qk.assemble(circuit_list,
                                       backend=self.backend,
                                       shots=self.shots,
                                       max_parallel_shots=1,
                                       max_parallel_experiments=0
                                       )

            job = self.backend.run(qobject_list)
            for counts in job.result().get_counts():
                outputs.append(self.sampler(counts))

        outputs = np.array(outputs).reshape(n_samples, -1)

        return self._scale_output(np.array(outputs))

    def grad(self, inputs, delta, samplewise=False):
        inputs = deepcopy(inputs)
        n_samples = inputs.shape[0]
        self.weight_partial = np.zeros((n_samples, *self.weight.shape))
        self.input_partial = np.zeros(
            (n_samples, self.n_features, self.n_targets))

        for i in range(self.n_weights_per_target):
            self.weight[i, :] += np.pi / 2
            self.weight_partial[:, i, :] = 1 / 2 * self(inputs)
            self.weight[i, :] += -np.pi
            self.weight_partial[:, i, :] += -1 / 2 * self(inputs)
            self.weight[i, :] += np.pi / 2

        weight_gradient = self.weight_partial * delta.reshape(n_samples, 1, -1)
        if not samplewise:
            weight_gradient = np.mean(weight_gradient, axis=0)

        if not self.last_layer:
            for i in range(self.n_features):
                inputs[:, i] += np.pi / 2
                self.input_partial[:, i, :] = 1 / 2 * self(inputs)
                inputs[:, i] += -np.pi
                self.input_partial[:, i, :] += -1 / 2 * self(inputs)
                inputs[:, i] += np.pi / 2

            delta = np.einsum("ij,ikj->ik", delta, self.input_partial)

        return weight_gradient, delta

    def randomize_weight(self):
        self.weight = np.random.uniform(
            -np.pi, np.pi, (self.n_weights_per_target, self.n_targets))

    def _scale_output(self, output):
        if self.scale == None:
            return output
        elif type(self.scale) != list:
            return self.scale * output
        else:
            a = self.scale[0]
            b = self.scale[1]
            return (b - a) * output + a


class Sigmoid():

    def __call__(self, x):
        x = 1 / (1 + np.exp(-x))

        return x

    def derivative(self, x):
        x = x * (1 - x)

        return x


class ReLu():

    def __call__(self, x):
        x = np.minimum(0, x)

        return x

    def derivative(self, x):
        x = x * (1 - x)

        return x


class Tanh():

    def __call__(self, x):
        x = np.tanh(x)

        return x

    def derivative(self, x):
        x = 1 - x**2

        return x


class Identity():

    def __call__(self, x):

        return x

    def derivative(self, x):

        return 1



from .utils import *
from .samplers import *
from .rco import *
from .parametrizations import *
from .optimizers import *
from .analysis import *
from .ansatzes import *
from .costfunction import *
from .data_encoders import *
from .layers import *
from .neuralnetwork import *



import numpy as np
import qiskit as qk
from copy import deepcopy
from tqdm.notebook import tqdm

from .optimizers import *
from .data_encoders import *
from .samplers import *
from .utils import *

"""
class Ansatz():
    def __init__(self, reps):
        self.reps = reps

    def __call__(self, circuit, data_register, weight):

        for i in range(self.reps):
            for j in range(self.n_qubits - 1):
                circuit.cx(data_register[j], data_register[j + 1])

            idx_start = i * self.n_qubits
            idx_end = (i + 1) * self.n_qubits
            for j, w in enumerate(weight[idx_start:idx_end]):
                circuit.ry(w, data_register[j])

        return circuit

    def calculate_n_weights(self, n_qubits):
        self.n_qubits = n_qubits
        self.n_weights_per_target = self.reps * self.n_qubits


class Ansatz2():
    def __call__(self, circuit, data_register, weight):
        n_qubits = data_register.size

        for i, w in enumerate(weight[:n_qubits // 2]):
            circuit.cx(data_register[2 * i], data_register[2 * i + 1])
            circuit.rz(w, data_register[i])
            circuit.cx(data_register[2 * i], data_register[2 * i + 1])

        for i, w in enumerate(weight[n_qubits // 2:]):
            circuit.cx(data_register[2 * i + 1], data_register[2 * i + 2])
            circuit.rz(w, data_register[i])
            circuit.cx(data_register[2 * i + 1], data_register[2 * i + 2])

        return circuit
"""


class Regularizer():
    def __call__(self, circuit, data_register, weight):
        n_qubits = data_register.size
        n_weights = weight.shape[0]

        for i, w in enumerate(weight[:n_weights / 2]):
            circuit.cx(data_register[2 * i], data_register[2 * i + 1])
            circuit.ry(w, data_register[2 * i])
            circuit.cx(data_register[2 * i], data_register[2 * i + 1])

        for i, w in enumerate(weight[n_weights / 2:]):
            circuit.cx(data_register[2 * i + 1], data_register[2 * i + 2])
            circuit.ry(w, data_register[2 * i + 1])
            circuit.cx(data_register[2 * i + 1], data_register[2 * i + 2])

        return circuit


class SwapTest():
    def __call__(self, circuit, register_a, register_b, ancilla_swap):
        circuit.h(ancilla_swap)
        for r_a, r_b in zip(register_a, register_b):
            circuit.cswap(ancilla_swap, r_a, r_b)

        circuit.h(ancilla_swap)

        return circuit


class ParallelModel():
    def __init__(self, n_features=None, n_targets=None, reps=1,  backend=None, shots=1000):
        self.parallel_encoder = ParallelEncoder()
        self.encoder = Encoder()
        self.ansatz = Ansatz()
        self.swap_test = SwapTest()

        self.n_features = n_features
        self.n_targets = n_targets
        self.reps = reps

        self.theta = np.random.uniform(
            0, 2 * np.pi, self.n_features * self.reps)
        self.backend = backend
        self.shots = shots

    def predict(self, x):
        n_samples, n_features = x.shape
        y_pred = []
        for i, x_ in enumerate(x):
            features = qk.QuantumRegister(self.n_features, name="features")
            predictions = qk.QuantumRegister(1, name="predictions")
            classical = qk.ClassicalRegister(1)
            registers = [features, predictions, classical]
            circuit = qk.QuantumCircuit(*registers)

            circuit = self.encoder(circuit, features, x_)
            for i in range(self.reps):
                start = i * self.n_features
                end = (i + 1) * self.n_features
                circuit = self.ansatz(circuit, features, self.theta[start:end])

            for i in range(n_features):
                circuit.cx(features[i], predictions)

            circuit.measure(predictions, classical)

            print(circuit)

            job = qk.execute(circuit, self.backend, shots=self.shots)
            counts = job.result().get_counts(circuit)
            """
            if "0" in counts:
                y_pred.append(
                    [2 * np.arccos(np.sqrt(counts["0"] / self.shots))])
            else:
                y_pred.append([np.pi])
            """
            if "0" in counts:
                y_pred.append(
                    [counts["0"] / self.shots])
            else:
                y_pred.append([0])

        return np.array(y_pred)

    def loss(self, x, y):
        n_samples, n_features = x.shape
        _, n_targets = y.shape
        n_ancilla = np.log2(n_samples)

        features = qk.QuantumRegister(n_features, name="features")
        predictions = qk.QuantumRegister(n_targets, name="predictions")
        ancilla_features = qk.QuantumRegister(
            n_ancilla, name="ancilla_feature")

        targets = qk.QuantumRegister(n_targets, name="targets")
        ancilla_targets = qk.QuantumRegister(n_ancilla, name="ancilla_target")

        ancilla_swap = qk.QuantumRegister(1, name="swap")
        classical = qk.ClassicalRegister(1)

        registers = [features, ancilla_features, predictions, targets,
                     ancilla_targets, ancilla_swap, classical]

        circuit = qk.QuantumCircuit(*registers)

        circuit = self.parallel_encoder(circuit, features, ancilla_features, x)
        circuit = self.parallel_encoder(circuit, targets, ancilla_targets, y)

        circuit.barrier()
        for i in range(self.reps):
            start = i * self.n_features
            end = (i + 1) * self.n_features
            circuit = self.ansatz(circuit, features, self.theta[start:end])

        circuit.barrier()

        for i in range(n_features):
            circuit.cx(features[i], predictions)

        register_a = predictions[:] + ancilla_features[:]
        register_b = targets[:] + ancilla_targets[:]
        circuit = self.swap_test(circuit, register_a, register_b, ancilla_swap)
        # circuit.cx(predictions, ancilla_swap)
        # circuit.cx(targets, ancilla_swap)
        circuit.measure(ancilla_swap, classical)

        # print(circuit)

        job = qk.execute(circuit, self.backend, shots=self.shots)
        counts = job.result().get_counts(circuit)
        if "0" in counts:
            loss = counts["0"] / self.shots
        else:
            loss = 0

        return loss

    def gradient(self, x, y):
        weight_gradient = np.zeros_like(self.theta)

        for i in range(len(self.theta)):
            self.theta[i] += np.pi / 2
            weight_gradient[i] += 1 / (2 * np.sqrt(2)) * self.loss(x, y)
            self.theta[i] += -np.pi
            weight_gradient[i] += -1 / (2 * np.sqrt(2)) * self.loss(x, y)
            self.theta[i] += np.pi / 2

        return weight_gradient


class RegularizedModel():
    def __init__(self, n_features=None, n_targets=None, reps=1, alpha=None, reverse=False, train_map=True,  backend=None, shots=1000, optimizer=None):
        self.encoder = RegularizedEncoder()
        self.ansatz = Ansatz()
        self.sampler = Parity()

        self.n_features = n_features
        self.n_targets = n_targets
        self.reps = reps
        self.alpha = alpha
        self.train_map = train_map
        self.backend = backend
        self.shots = shots
        self.optimizer = optimizer

        self.loss = []
        self.map_params = []

        self.n_qubits = n_features + 1
        self.theta = np.random.uniform(
            0, 2 * np.pi, self.n_qubits * self.reps + self.n_features)

        if self.train_map:
            self.theta[:self.n_features] = 0
        else:
            self.theta[:self.n_features] = np.pi / 2

        if not self.optimizer == None:
            self.optimizer.initialize(self.theta.shape)

    def predict(self, x):
        outputs = []
        circuit_list = []
        for i, x_ in enumerate(x):
            data_register = qk.QuantumRegister(self.n_qubits)
            classical = qk.ClassicalRegister(self.n_qubits)
            registers = [data_register, classical]
            circuit = qk.QuantumCircuit(*registers)

            circuit = self.encoder(
                circuit, data_register, x_, self.theta[:self.n_features])

            for j in range(self.reps):
                start = j * self.n_qubits + self.n_features
                end = (j + 1) * self.n_qubits + self.n_features
                circuit = self.ansatz(
                    circuit, data_register, self.theta[start:end])

            circuit.measure(data_register, classical)
            circuit_list.append(circuit)

        transpiled_list = qk.transpile(circuit_list, backend=self.backend)
        qobject_list = qk.assemble(transpiled_list,
                                   backend=self.backend,
                                   shots=self.shots,
                                   max_parallel_shots=1,
                                   max_parallel_experiments=0
                                   )
        job = self.backend.run(qobject_list)

        for circuit in circuit_list:
            counts = job.result().get_counts(circuit)
            outputs.append(self.sampler(counts))

        return np.array(outputs).reshape(-1, 1)

    def gradient(self, x, y):
        n_samples = x.shape[0]
        y_pred = self.predict(x)
        delta = (y_pred - y)
        weight_gradient = np.zeros((n_samples, len(self.theta)))

        if self.train_map:
            start = 0
        else:
            start = self.n_features

        for i in range(start, len(self.theta)):

            self.theta[i] += np.pi / 2
            weight_gradient[:, i] += 1 / \
                (2 * np.sqrt(2)) * self.predict(x)[:, 0]
            self.theta[i] += -np.pi
            weight_gradient[:, i] += -1 / \
                (2 * np.sqrt(2)) * self.predict(x)[:, 0]
            self.theta[i] += np.pi / 2

        weight_gradient = weight_gradient * delta
        weight_gradient = np.mean(weight_gradient, axis=0)

        return weight_gradient

    def train(self, x, y, epochs=100, verbose=False):
        if verbose:
            dec = tqdm
        else:
            dec = identity

        for i in dec(range(epochs)):
            y_pred = self.predict(x)
            self.loss.append(np.mean((y_pred - y)**2))
            self.map_params.append(deepcopy(self.theta[:self.n_features]))

            gradient = self.gradient(x, y)
            gradient_mod = self.optimizer([gradient])[0]

            penalty = np.zeros_like(gradient)
            if self.train_map:
                penalty[:self.n_features] = self.theta[:self.n_features]

            self.theta -= self.optimizer.lr * gradient_mod + self.alpha * penalty

            if verbose:
                print(f"epoch: {i}, loss: {self.loss[-1]}")

        y_pred = self.predict(x)
        self.loss.append(np.mean((y_pred - y)**2))
        self.map_params.append(self.theta[:self.n_features])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

# Package meta-data.
NAME = "qcnpy"
DESCRIPTION = "Variational quantum algorithms for machine learning."
URL = "https://github.com/KristianWold/qcnpy"
EMAIL = "kristian.wold@hotmail.com"
AUTHOR = "Kristian Wold"
REQUIRES_PYTHON = '>=3.8.0'

REQUIRES_INSTALL = [
    "numpy",
    "matplotlib",
    "qiskit",
    "tqdm",
]

REQUIRES_EXTRAS = {
    "dev": [
        "pytest",
        "pytest-cov",
        "flake8>=3.9.2",
        "isort",
        "twine",
    ],
}


with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

about = {}
with open(os.path.join(here, NAME, "__version__.py")) as f:
    exec(f.read(), about)

VERSION = about['__version__']


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    packages=find_packages(exclude=["tests", ]),
    python_requires=REQUIRES_PYTHON,
    install_requires=REQUIRES_INSTALL,
    extras_require=REQUIRES_EXTRAS,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import numpy as np
import pickle
import os


PROJECT_ROOT_DIR = "../../results"
DATA_ID = "../../results/data"
FIGURE_ID = "../../results/figures"

if not os.path.exists(PROJECT_ROOT_DIR):
    os.mkdir(PROJECT_ROOT_DIR)
if not os.path.exists(FIGURE_ID):
    os.makedirs(FIGURE_ID)
if not os.path.exists(DATA_ID):
    os.makedirs(DATA_ID)


def image_path(fig_id):
    return os.path.join(FIGURE_ID, fig_id)


def data_path(data_id):
    return os.path.join(DATA_ID, data_id)


def identity(func):
    return func


def saver(object, filename):
    pickle.dump(object, open(filename, "wb"))


def loader(filename):
    object = pickle.load(open(filename, "rb"))

    return object


def unpack_list(list_):
    list_flat = []
    for l in list_:
        list_flat.append(l.flatten())

    list_flat = np.concatenate(list_flat).reshape(-1, 1)

    return list_flat


def gaussian(x, mean, var):
    if type(mean) == float:
        mean = np.array([[mean]])

    if type(var) == float:
        var = np.array([[var]])

    var_inv = np.linalg.inv(var)
    diag = np.diag((x - mean) @ var_inv @ (x - mean).T).reshape(-1, 1)
    y = np.exp(-0.5 * diag)

    return y


def random_mixed_gaussian(x, n_gaussians=3):
    dim = x.shape[-1]
    mean = np.random.uniform(0, 1, (n_gaussians, 1, dim))

    var = []
    for i in range(n_gaussians):
        var_ = np.random.uniform(-0.001, 0.001, (dim, dim))
        var_[np.diag_indices(dim)] = np.random.uniform(0.005, 0.05, dim)
        var.append(var_)

    alpha = np.random.uniform(-1, 1, n_gaussians)

    y = 0
    for i in range(n_gaussians):
        y += alpha[i] * gaussian(x, mean[i], var[i])

    return y


def scaler(x, mode="uniform", a=0, b=np.pi):
    if mode == "uniform":
        x = x - np.min(x, axis=0)
        x = (b - a) * x / np.max(x, axis=0)
        x = x + a
    if mode == "standard":
        x = (x - np.mean(x, axis=0)) / np.std(x, axis=0)

    return x


def generate_meshgrid(x):

    x = np.meshgrid(*x)
    x = [np.ravel(x_).reshape(-1, 1) for x_ in x]
    x = np.hstack(x)

    return x


def r2(models, x, y):
    y_avg = np.mean(y)

    if type(models) != list:
        models = [models]

    r2_scores = []
    for model in models:
        y_pred = model.predict(x)
        r2 = 1 - np.mean((y_pred - y)**2) / np.mean((y - y_avg)**2)
        r2_scores.append(r2)

    return sum(r2_scores) / len(r2_scores)



import numpy as np
import qiskit as qk
from math import ceil
from math import floor


class Encoder():
    """Quantum circuit for encoding data by Pauli rotations.

    Parameters
    ----------
    mode : str
        Specify what Pauli rotations to use. "x", "y" and "z" corresponds
        to Rx, Ry and Rz rotation, respectivly.
    """

    def __init__(self, mode="y"):
        self.mode = mode

    def __call__(self, circuit, data_register, weight, data):
        """Apply gates to circuit for encoding data.

        Parameters
        ----------
        mode : str
            Specify what Pauli rotations to use. "x", "y" and "z" corresponds
            to Rx, Ry and Rz rotation, respectivly.
        """

        n_qubits = data_register.size
        n_features = data.shape[0]

        for i, x in enumerate(data):
            if self.mode == "x":
                circuit.rx(x, data_register[i])
            if self.mode == "y":
                circuit.ry(x, data_register[i])
            if self.mode == "z":
                circuit.h(data_register[i])
                circuit.rz(x, data_register[i])

        # apply Hadamard on latent qubits
        if n_qubits > n_features:
            for i in range(n_features, n_qubits):
                circuit.h(data_register[i])

        return circuit

    def calculate_n_weights(self, n_qubits):
        self.n_qubits = n_qubits
        self.n_weights_per_target = 0


class RegularizedEncoder():
    def __init__(self, reg=True):
        self.reg = reg

    def __call__(self, circuit, data_register, data, params):
        n_qubits = data_register.size
        n_features = data.shape[0]
        n_params = params.shape[0]

        for i, x in enumerate(data):
            circuit.ry(x, data_register[i])

        if n_qubits > n_features:
            for i in range(n_features, n_qubits):
                circuit.h(data_register[i])

        for i, w in enumerate(params[:n_params // 2]):
            circuit.cx(data_register[2 * i], data_register[2 * i + 1])
            circuit.ry(w, data_register[2 * i])
            circuit.cx(data_register[2 * i], data_register[2 * i + 1])

        for i, w in enumerate(params[n_params // 2:]):
            circuit.cx(data_register[2 * i + 1], data_register[2 * i + 2])
            circuit.ry(w, data_register[2 * i + 1])
            circuit.cx(data_register[2 * i + 1], data_register[2 * i + 2])

        if self.reg:
            for i, x in enumerate(data):
                circuit.ry(-x, data_register[i])
        else:
            for i, x in enumerate(data):
                circuit.ry(x, data_register[i])

        if n_qubits > n_features:
            for i in range(n_features, n_qubits):
                circuit.h(data_register[i])

        return circuit


class ParallelEncoder():
    def __call__(self, circuit, data_register, ancilla, data):
        n_samples, n_features = data.shape
        n_ancilla = ancilla.size

        binary_ref = n_ancilla * [0]
        for i in range(n_samples):
            binary = interger_to_binary(i, n_ancilla)

            for j, (b, b_ref) in enumerate(zip(binary, binary_ref)):
                if b != b_ref:
                    circuit.x(ancilla[j])

            for j in range(n_features):
                circuit.cry(data[i, j], ancilla,
                            data_register[j])

            binary_ref = binary

        circuit.x(ancilla)

        for i in range(n_features - 1):
            circuit.cx(data_register[i], data_register[i + 1])

        return circuit


def amplitude_encoding(data, circuit, storage, inverse=False):
    N = data.shape[0]
    n = int(np.log2(N))

    circuit.ry(calculate_rotation(data, 0, 0), storage[0])
    for i in range(1, n):
        binary_ref = i * [0]
        circuit.x(storage[:i])

        for j in range(2**i):
            binary = interger_to_binary(j, i)

            for k, (b, b_ref) in enumerate(zip(binary, binary_ref)):
                if b != b_ref:
                    circuit.x(storage[k])

            circuit.mcry(calculate_rotation(data, i, j),
                         storage[:i], storage[i], q_ancillae=None)
            binary_ref = binary

    if inverse:
        circuit = circuit.inverse()

    return circuit


def basis_encoding(x):
    M, N = x.shape

    clas_reg = qk.ClassicalRegister(N)
    loading_reg = qk.QuantumRegister(N, name="loading")
    storage_reg = qk.QuantumRegister(N, name="storage")
    ancillas = qk.QuantumRegister(N, name="ancillas")
    branches = qk.QuantumRegister(2, name="branches")

    circuit = qk.QuantumCircuit(
        clas_reg, loading_reg, storage_reg, ancillas, branches)
    circuit.x(branches[1])

    for i in range(M):
        for j in range(N):
            if x[i, j] == 1:
                circuit.x(loading_reg[j])
                circuit.cx(branches[1], storage_reg[j])

        circuit.cx(branches[1], branches[0])
        theta = -1 / np.sqrt(M - i)
        circuit.cry(2 * np.arcsin(theta), branches[0], branches[1])

        circuit.toffoli(loading_reg, storage_reg, ancillas)
        circuit.x(loading_reg)
        circuit.x(storage_reg)
        circuit.toffoli(loading_reg, storage_reg, ancillas)

        circuit.mcx(ancillas, branches[0])

        circuit.toffoli(loading_reg, storage_reg, ancillas)
        circuit.x(loading_reg)
        circuit.x(storage_reg)
        circuit.toffoli(loading_reg, storage_reg, ancillas)

        for j in range(N):
            if x[i, j] == 1:
                circuit.x(loading_reg[j])
                circuit.cx(branches[1], storage_reg[j])

    circuit.measure(storage_reg, clas_reg)
    return circuit


# helper functions
################################################


def calculate_rotation(data, i, j):
    n = int(np.log2(len(data)))

    idx1 = (2 * j + 1) * 2**(n - i - 1)
    idx2 = (j + 1) * 2**(n - i)
    idx3 = j * 2**(n - i)
    idx4 = (j + 1) * 2**(n - i)

    if i == n - 1:
        a1 = data[idx1]
    else:
        a1 = np.sqrt(np.sum(np.abs(data[idx1:idx2])**2))

    a2 = np.sqrt(np.sum(np.abs(data[idx3:idx4])**2))
    if a2 == 0:
        return 0
    else:
        return 2 * np.arcsin(a1 / a2)


def interger_to_binary(integer, digits):
    binary = [int(b) for b in bin(integer)[2:]]
    binary = (digits - len(binary)) * [0] + binary
    return binary


def float_to_binary(x, digits=4):
    binary_rep = []
    if x > 0:
        binary_rep.append(0)
    else:
        binary_rep.append(1)

    x = abs(x)

    for i in range(digits):
        digit = floor(x * 2**(i))
        binary_rep.append(digit)
        x -= digit * 2**(-i)

    return binary_rep


def design_matrix_to_binary(X, digits=4):
    M, N = X.shape
    X_ = []

    for row in X:
        row_ = [float_to_binary(feature, digits) for feature in row]
        X_.append(row_)

    return np.array(X_).reshape(M, (digits + 1) * N)
